from enum import Enum

from pydantic import BaseModel


class PredictionBoundingBoxXYWH(BaseModel):
    """
    Represents a bounding box prediction, defined by its X and Y coordinates,
    width, height, and confidence score.

    Attributes:
    - x (float | int): The X coordinate of the bounding box.
    - y (float | int): The Y coordinate of the bounding box.
    - width (float | int): The width of the bounding box.
    - height (float | int): The height of the bounding box.
    - confidence (float | int): The confidence score of the bounding box prediction.
    """

    x: float | int
    y: float | int
    width: float | int
    height: float | int
    confidence: float | int


class PredictionBoundingBoxXYXY(BaseModel):
    """
    Represents a bounding box prediction, defined by its minimum and maximum X and Y coordinates,
    along with a confidence score.

    Attributes:
    - xmin (float | int): The minimum X coordinate of the bounding box.
    - xmax (float | int): The maximum X coordinate of the bounding box.
    - ymin (float | int): The minimum Y coordinate of the bounding box.
    - ymax (float | int): The maximum Y coordinate of the bounding box.
    - confidence (float | int): The confidence score of the bounding box prediction.
    """

    xmin: float | int
    xmax: float | int
    ymin: float | int
    ymax: float | int
    confidence: float | int


class YoloV7PoseKeypoint(BaseModel):
    """
    Represents a YOLOv7 pose keypoint with floating point x and y coordinates and a
    confidence score.

    Attributes:
        x (float | int): The x-coordinate of the keypoint.
        y (float | int): The y-coordinate of the keypoint.
        conf (float | int): The confidence score of the keypoint.
    """

    x: float | int
    y: float | int
    conf: float | int


class YoloV7PoseJoints(Enum):
    """
    Enumerates the different joints that can be detected by a YOLOv7 Pose model.
    """

    NOSE = 7
    LEFT_EYE = 10
    RIGHT_EYE = 13
    LEFT_EAR = 16
    RIGHT_EAR = 19
    LEFT_SHOULDER = 22
    RIGHT_SHOULDER = 25
    LEFT_ELBOW = 28
    RIGHT_ELBOW = 31
    LEFT_WRIST = 34
    RIGHT_WRIST = 37
    LEFT_HIP = 40
    RIGHT_HIP = 43
    LEFT_KNEE = 46
    RIGHT_KNEE = 49
    LEFT_ANKLE = 52
    RIGHT_ANKLE = 55


class YoloV7PoseKeypointsIndex(Enum):
    """
    Enumerates the different keypoint indices that correspond to the various joints detected by a YOLOv7 Pose model.
    """

    BATCH_ID = 0
    CLASS_ID = 1
    X = 2
    Y = 3
    WIDTH = 4
    HEIGHT = 5
    CONFIDENCE = 6
    X_NOSE = 7
    Y_NOSE = 8
    CONFIDENCE_NOSE = 9
    X_LEFT_EYE = 10
    Y_LEFT_EYE = 11
    CONFIDENCE_LEFT_EYE = 12
    X_RIGHT_EYE = 13
    Y_RIGHT_EYE = 14
    CONFIDENCE_RIGHT_EYE = 15
    X_LEFT_EAR = 16
    Y_LEFT_EAR = 17
    CONFIDENCE_LEFT_EAR = 18
    X_RIGHT_EAR = 19
    Y_RIGHT_EAR = 20
    CONFIDENCE_RIGHT_EAR = 21
    X_LEFT_SHOULDER = 22
    Y_LEFT_SHOULDER = 23
    CONFIDENCE_LEFT_SHOULDER = 24
    X_RIGHT_SHOULDER = 25
    Y_RIGHT_SHOULDER = 26
    CONFIDENCE_RIGHT_SHOULDER = 27
    X_LEFT_ELBOW = 28
    Y_LEFT_ELBOW = 29
    CONFIDENCE_LEFT_ELBOW = 30
    X_RIGHT_ELBOW = 31
    Y_RIGHT_ELBOW = 32
    CONFIDENCE_RIGHT_ELBOW = 33
    X_LEFT_WRIST = 34
    Y_LEFT_WRIST = 35
    CONFIDENCE_LEFT_WRIST = 36
    X_RIGHT_WRIST = 37
    Y_RIGHT_WRIST = 38
    CONFIDENCE_RIGHT_WRIST = 39
    X_LEFT_HIP = 40
    Y_LEFT_HIP = 41
    CONFIDENCE_LEFT_HIP = 42
    X_RIGHT_HIP = 43
    Y_RIGHT_HIP = 44
    CONFIDENCE_RIGHT_HIP = 45
    X_LEFT_KNEE = 46
    Y_LEFT_KNEE = 47
    CONFIDENCE_LEFT_KNEE = 48
    X_RIGHT_KNEE = 49
    Y_RIGHT_KNEE = 50
    CONFIDENCE_RIGHT_KNEE = 51
    X_LEFT_ANKLE = 52
    Y_LEFT_ANKLE = 53
    CONFIDENCE_LEFT_ANKLE = 54
    X_RIGHT_ANKLE = 55
    Y_RIGHT_ANKLE = 56
    CONFIDENCE_RIGHT_ANKLE = 57



from enum import Enum
from typing import List

import numpy as np
import torch
from numpy import ndarray
from torchvision import transforms

from aipose.downloader import Downloader
from aipose.models.yolov7.types import (
    PredictionBoundingBoxXYWH,
    PredictionBoundingBoxXYXY,
    YoloV7PoseJoints,
    YoloV7PoseKeypoint,
    YoloV7PoseKeypointsIndex,
)
from aipose.utils import letterbox, non_max_suppression_kpt, output_to_keypoint


class YoloV7PoseKeypoints:
    """
    A class that represents pose keypoints and their confidences extracted by the YOLOv7 model.
    """

    _step_keypoint: int = 3
    raw_keypoints: List[float]

    def __init__(
        self,
        raw_keypoints: List[float],
        height_prediction: int,
        width_prediction: int,
        height_original: int,
        width_original: int,
    ):
        """
        Initialize a YoloV7PoseKeypoints instance.

        :param raw_keypoints: A list of raw keypoints and their confidences.
        :param height: The height of the image.
        :param width: The width of the image.
        """
        self.raw_keypoints = raw_keypoints
        self.raw_keypoints[
            YoloV7PoseKeypointsIndex.WIDTH.value
        ] = self._convert_dimension(
            self.raw_keypoints[YoloV7PoseKeypointsIndex.WIDTH.value],
            width_prediction,
            width_original,
        )
        self.raw_keypoints[
            YoloV7PoseKeypointsIndex.HEIGHT.value
        ] = self._convert_dimension(
            self.raw_keypoints[YoloV7PoseKeypointsIndex.HEIGHT.value],
            width_prediction,
            width_original,
        )
        self.raw_keypoints[
            YoloV7PoseKeypointsIndex.X.value : YoloV7PoseKeypointsIndex.Y.value + 1
        ] = self._convert_point(
            self.raw_keypoints[YoloV7PoseKeypointsIndex.X.value],
            self.raw_keypoints[YoloV7PoseKeypointsIndex.Y.value],
            width_prediction,
            height_prediction,
            width_original,
            height_original,
        )
        self.raw_keypoints[7:] = self._convert_keypoints_with_confidence(
            self.raw_keypoints[7:],
            width_prediction,
            height_prediction,
            width_original,
            height_original,
        )

    @property
    def batch_id(self) -> int:
        return self.raw_keypoints[0]

    @property
    def class_id(self) -> int:
        return self.raw_keypoints[1]

    @property
    def body_keypoints(self) -> List[float]:
        return self.raw_keypoints[7:]

    @property
    def raw_keypoint(self) -> List[float]:
        return self.raw_keypoints

    @property
    def human_confidence(self) -> List[float]:
        return self.raw_keypoints[YoloV7PoseKeypointsIndex.CONFIDENCE]

    @property
    def human_center(self) -> List[float]:
        x = self.raw_keypoints[YoloV7PoseKeypointsIndex.X.value]
        y = self.raw_keypoints[YoloV7PoseKeypointsIndex.Y.value]
        return (x, y)

    @property
    def prediction_bounding_box_xyxy(self) -> PredictionBoundingBoxXYXY:
        points = self.keypoints_coordinates
        xmin = min([x for x, y in points])
        xmax = max([x for x, y in points])
        ymin = min([y for x, y in points])
        ymax = max([y for x, y in points])
        confidence = self.raw_keypoints[6]

        return PredictionBoundingBoxXYXY(
            xmin=xmin, xmax=xmax, ymin=ymin, ymax=ymax, confidence=confidence
        )

    @property
    def prediction_bounding_box_xywh(self) -> PredictionBoundingBoxXYWH:
        return PredictionBoundingBoxXYWH(
            x=self.raw_keypoints[2],
            y=self.raw_keypoints[3],
            width=self.raw_keypoints[4],
            height=self.raw_keypoints[5],
            confidence=self.raw_keypoints[6],
        )

    @property
    def is_backwards(self) -> bool:
        return (
            self.get_keypoint(YoloV7PoseJoints.LEFT_EAR).x
            < self.get_keypoint(YoloV7PoseJoints.RIGHT_EAR).x
        )

    @property
    def keypoints_coordinates(self) -> ndarray:
        body_keypoints = self.body_keypoints
        points = np.array([*zip(body_keypoints[::3], body_keypoints[1::3])])
        return points

    def total_confidence_over(
        self, expected_confidence: float
    ) -> List[YoloV7PoseJoints]:
        """
        This method returns a list of YoloV7PoseJoints objects where the confidence level is greater than a given expected confidence level.


        Attributes:
            expected_confidence (float): the confidence level threshold to filter keypoints

        Returns:
            List[YoloV7PoseJoints]: a list of YoloV7PoseJoints objects where the confidence level is greater than the expected confidence level.
        """
        keypoints_over_confidence = []

        for pose_joint_index in YoloV7PoseJoints:
            keypoint = self.get_keypoint(pose_joint_index)
            if keypoint.conf > expected_confidence:
                keypoints_over_confidence.append(keypoint)

        return keypoints_over_confidence

    def get_keypoint(self, keypoint: YoloV7PoseJoints | int) -> YoloV7PoseKeypoint:
        if isinstance(keypoint, Enum):
            keypoint = keypoint.value
        return self._get_x_y_conf(keypoint)

    def cosine_similarity(self, pose: "YoloV7PoseKeypoints") -> float:
        a = self.get_points_normalize_by_bbox()
        b = pose.get_points_normalize_by_bbox()

        return round(np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b)), 4)

    def get_points_normalize_by_bbox(self) -> ndarray:
        bbox = self.prediction_bounding_box_xywh
        xmin, ymin, width, height = bbox.x, bbox.y, bbox.width, bbox.height

        body_keypoints = self.body_keypoints
        body_keypoints_normalized = np.zeros_like(body_keypoints)

        for i in range(0, len(body_keypoints), 3):
            x, y, confidence = (
                body_keypoints[i],
                body_keypoints[i + 1],
                body_keypoints[i + 2],
            )
            x_normalized = (x - xmin) / width
            y_normalized = (y - ymin) / height
            body_keypoints_normalized[i] = x_normalized
            body_keypoints_normalized[i + 1] = y_normalized
            body_keypoints_normalized[i + 2] = confidence

        return body_keypoints_normalized

    def turn(self) -> "YoloV7PoseKeypoints":
        self._swap_keypoints(YoloV7PoseJoints.LEFT_EYE, YoloV7PoseJoints.RIGHT_EYE)
        self._swap_keypoints(YoloV7PoseJoints.LEFT_EAR, YoloV7PoseJoints.RIGHT_EAR)
        self._swap_keypoints(
            YoloV7PoseJoints.LEFT_SHOULDER, YoloV7PoseJoints.RIGHT_SHOULDER
        )
        self._swap_keypoints(YoloV7PoseJoints.LEFT_ELBOW, YoloV7PoseJoints.RIGHT_ELBOW)
        self._swap_keypoints(YoloV7PoseJoints.LEFT_WRIST, YoloV7PoseJoints.RIGHT_WRIST)
        self._swap_keypoints(YoloV7PoseJoints.LEFT_HIP, YoloV7PoseJoints.RIGHT_HIP)
        self._swap_keypoints(YoloV7PoseJoints.LEFT_KNEE, YoloV7PoseJoints.RIGHT_KNEE)
        self._swap_keypoints(YoloV7PoseJoints.LEFT_ANKLE, YoloV7PoseJoints.RIGHT_ANKLE)
        return self

    def _swap_keypoints(
        self, keypoint_a: YoloV7PoseJoints | int, keypoint_b: YoloV7PoseJoints | int
    ):
        """
        Swap the values of two keypoints.

        :param a_index: The index of the first keypoint.
        :param b_index: The index of the second keypoint.
        """
        if isinstance(keypoint_a, Enum):
            a_index = keypoint_a.value

        if isinstance(keypoint_b, Enum):
            b_index = keypoint_b.value

        end_index_a = a_index + self._step_keypoint
        a_x = self.raw_keypoints[a_index:end_index_a][0]
        a_y = self.raw_keypoints[a_index:end_index_a][1]
        a_conf = self.raw_keypoints[a_index:end_index_a][2]

        end_index_b = b_index + self._step_keypoint
        b_x = self.raw_keypoints[b_index:end_index_b][0]
        b_y = self.raw_keypoints[b_index:end_index_b][1]
        b_conf = self.raw_keypoints[b_index:end_index_b][2]

        self.raw_keypoints[a_index:end_index_a][0] = b_x
        self.raw_keypoints[a_index:end_index_a][1] = b_y
        self.raw_keypoints[a_index:end_index_a][2] = b_conf

        self.raw_keypoints[b_index:end_index_b][0] = a_x
        self.raw_keypoints[b_index:end_index_b][1] = a_y
        self.raw_keypoints[b_index:end_index_b][2] = a_conf

    def _convert_dimension(self, value, original_dim, target_dim):
        converted_value = (value / original_dim) * target_dim
        return converted_value

    def _convert_point(self, x, y, width1, height1, width2, height2):
        new_x = (x / width1) * width2
        new_y = (y / height1) * height2
        return [new_x, new_y]

    def _convert_keypoints_with_confidence(
        self, keypoints_with_confidence, width1, height1, width2, height2
    ):
        converted_keypoints = []

        for i in range(0, len(keypoints_with_confidence), 3):
            x = keypoints_with_confidence[i]
            y = keypoints_with_confidence[i + 1]

            converted_point = self._convert_point(
                x, y, width1, height1, width2, height2
            )
            confidence = keypoints_with_confidence[i + 2]

            converted_keypoints.extend(converted_point + [confidence])

        return converted_keypoints

    def _get_x_y_conf(self, start_index: int) -> YoloV7PoseKeypoint:
        """
        Get the x, y, and confidence values for a single keypoint.

        :param start_index: The index at which to start reading values from the raw_keypoints list.
        :return: A YoloV7PoseKeypoint object representing a single keypoint.
        """
        end_index = start_index + self._step_keypoint
        x = self.raw_keypoints[start_index:end_index][0]
        y = self.raw_keypoints[start_index:end_index][1]
        conf = self.raw_keypoints[start_index:end_index][2]
        return YoloV7PoseKeypoint(x=x, y=y, conf=conf)

    def __str__(self) -> str:
        return str(self.raw_keypoints)

    def __repr__(self) -> str:
        return str(self.raw_keypoints)


class YoloV7Pose:
    aipose_model_path: str = ""
    _model_repo: str = "WongKinYiu/yolov7"
    aipose_model_hash: str = "62ca91ec6612b22bef0ab4c95f3e2d07"
    aipose_model_file_name: str = "yolov7-w6-pose.pt"
    model_url_download: str = "https://huggingface.co/Tlaloc-Es/yolov7-w6-pose.pt/resolve/main/yolov7-w6-pose.pt"  # noqa: E501

    def __init__(self):
        self.aipose_model_path = Downloader().download(
            self.aipose_model_file_name, self.aipose_model_hash, self.model_url_download
        )

        self.device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

        try:
            self.model = torch.hub.load(
                self._model_repo,
                "custom",
                f"{ self.aipose_model_path}",
                trust_repo=True,
            )
        except Exception as e:  # noqa: F841
            weigths = torch.load(self.aipose_model_path, map_location=self.device)

        self.model = weigths["model"]
        self.model.float().eval()

        if torch.cuda.is_available():
            self.model.half().to(self.device)

    def __call__(
        self, image: ndarray, conf_thres=0.25, iou_thres=0.65
    ) -> List[YoloV7PoseKeypoints]:
        # Resize and pad image
        image_letterbox = letterbox(image, 960, stride=64, auto=True)[
            0
        ]  # shape: (567, 960, 3)
        # Apply transforms
        size_image = image.shape
        size_prediction = image_letterbox.shape

        image_letterbox = transforms.ToTensor()(
            image_letterbox
        )  # torch.Size([3, 567, 960])
        if torch.cuda.is_available():
            image_letterbox = image_letterbox.half().to(self.device)
        # Turn image into batch
        image_letterbox = image_letterbox.unsqueeze(0)  # torch.Size([1, 3, 567, 960])
        with torch.no_grad():
            original_output, _ = self.model(image_letterbox)

        original_output = non_max_suppression_kpt(
            original_output,
            conf_thres,
            iou_thres,
            nc=self.model.yaml["nc"],
            nkpt=self.model.yaml["nkpt"],
            kpt_label=True,
        )

        with torch.no_grad():
            output = output_to_keypoint(original_output)

        return [
            YoloV7PoseKeypoints(
                prediction,
                size_prediction[0],
                size_prediction[1],
                size_image[0],
                size_image[1],
            )
            for prediction in output
        ]



from numpy import ndarray


class FrameManagerBase:
    """
    A base class for managing video frames.

    Attributes:
        stop (bool): A boolean flag indicating if frame reading should be stopped.

    Methods:
        __init__(): Initializes a new instance of the FrameManagerBase class.
        before_read_frame(): Method called before reading a new frame.
        frame_received(frame: ndarray) -> ndarray: Method called for each frame.
        to_stop() -> bool: Returns the current value of the stop flag.
        stream_started(source: str | int) -> None: Method called when a new stream is started.
        stream_ended() -> None: Method called when the video stream ends.
    """

    stop = False

    def __init__(self):
        """
        Initializes a new instance of the FrameManagerBase class.
        """

    def before_read_frame(self):
        """
        Method called before reading a new frame.
        """

    def frame_received(self, frame: ndarray) -> ndarray:
        """
        Method called for each frame.
        Args:
            frame (ndarray): A numpy array representing the frame.

        Returns:
            ndarray: A numpy array representing the modified frame.
        """
        return frame

    def to_stop(self) -> bool:
        """
        Returns the current value of the stop flag.
        Returns:
            bool: The current value of the stop flag.
        """
        return self.stop

    def stream_started(self, source: str | int) -> None:
        """
        Method called when a new video stream is started.
        Args:
            source (str | int): The source of the video stream (either a string path or
                an integer camera ID or a rtsp uri).
        """
        pass

    def stream_ended(self) -> None:
        """
        Method called when the video stream ends.
        """
        pass



import numpy as np


def apply_gamma_correction(original_image, gamma=2.2):
    """
    Apply gamma correction to an image.

    Args:
        original_image (numpy.ndarray): The input image.
        gamma (float): The gamma value for correction (default is 2.2).

    Returns:
        numpy.ndarray: The gamma-corrected image.
    """
    # Normalize the image by dividing all pixel values by 255.0
    image_normalized = original_image / 255.0

    # Apply gamma correction to the normalized image
    corrected_image = np.power(image_normalized, gamma)

    # Scale the corrected values back to the range [0, 255] and convert to uint8
    corrected_image = (corrected_image * 255).astype(np.uint8)

    return corrected_image



import logging

import click

from aipose.frame import FramePlot
from aipose.models.yolov7.domain import YoloV7Pose
from aipose.stream import process_webcam

logging.basicConfig(level=logging.INFO)


@click.command()
def webcam():
    model = YoloV7Pose()
    frame_plot = FramePlot(model)
    process_webcam(frame_plot)


if __name__ == "__main__":
    webcam()



import cv2
import matplotlib.pyplot as plt
import numpy as np
from numpy import ndarray
from torch import Tensor


def plot_skeleton_kpts(im, kpts, steps, orig_shape=None):
    # Plot the skeleton and keypointsfor coco datatset
    palette = np.array(
        [
            [255, 128, 0],
            [255, 153, 51],
            [255, 178, 102],
            [230, 230, 0],
            [255, 153, 255],
            [153, 204, 255],
            [255, 102, 255],
            [255, 51, 255],
            [102, 178, 255],
            [51, 153, 255],
            [255, 153, 153],
            [255, 102, 102],
            [255, 51, 51],
            [153, 255, 153],
            [102, 255, 102],
            [51, 255, 51],
            [0, 255, 0],
            [0, 0, 255],
            [255, 0, 0],
            [255, 255, 255],
        ]
    )

    skeleton = [
        [16, 14],
        [14, 12],
        [17, 15],
        [15, 13],
        [12, 13],
        [6, 12],
        [7, 13],
        [6, 7],
        [6, 8],
        [7, 9],
        [8, 10],
        [9, 11],
        [2, 3],
        [1, 2],
        [1, 3],
        [2, 4],
        [3, 5],
        [4, 6],
        [5, 7],
    ]

    pose_limb_color = palette[
        [9, 9, 9, 9, 7, 7, 7, 0, 0, 0, 0, 0, 16, 16, 16, 16, 16, 16, 16]
    ]
    pose_kpt_color = palette[[16, 16, 16, 16, 16, 0, 0, 0, 0, 0, 0, 9, 9, 9, 9, 9, 9]]
    radius = 5
    num_kpts = len(kpts) // steps

    for kid in range(num_kpts):
        r, g, b = pose_kpt_color[kid]
        x_coord, y_coord = kpts[steps * kid], kpts[steps * kid + 1]
        if not (x_coord % 640 == 0 or y_coord % 640 == 0):
            if steps == 3:
                conf = kpts[steps * kid + 2]
                if conf < 0.5:
                    continue
            cv2.circle(
                im, (int(x_coord), int(y_coord)), radius, (int(r), int(g), int(b)), -1
            )
            cv2.putText(
                im,
                str(kid),
                (int(x_coord) - 5, int(y_coord) + 5),
                cv2.FONT_HERSHEY_SIMPLEX,
                0.7,
                (246, 255, 12),
                3,
            )

    for sk_id, sk in enumerate(skeleton):
        r, g, b = pose_limb_color[sk_id]
        pos1 = (int(kpts[(sk[0] - 1) * steps]), int(kpts[(sk[0] - 1) * steps + 1]))
        pos2 = (int(kpts[(sk[1] - 1) * steps]), int(kpts[(sk[1] - 1) * steps + 1]))
        if steps == 3:
            conf1 = kpts[(sk[0] - 1) * steps + 2]
            conf2 = kpts[(sk[1] - 1) * steps + 2]
            if conf1 < 0.5 or conf2 < 0.5:
                continue
        if pos1[0] % 640 == 0 or pos1[1] % 640 == 0 or pos1[0] < 0 or pos1[1] < 0:
            continue
        if pos2[0] % 640 == 0 or pos2[1] % 640 == 0 or pos2[0] < 0 or pos2[1] < 0:
            continue
        cv2.line(im, pos1, pos2, (int(r), int(g), int(b)), thickness=2)


def plot(image: np.ndarray, output: np.ndarray, plot_image=True, return_img=False):
    for idx in range(output.shape[0]):
        plot_skeleton_kpts(image, output[idx, 7:].T, 3)

    if plot_image:
        plt.figure(figsize=(8, 8))
        plt.axis("off")
        plt.imshow(image)
        plt.show()

    if return_img:
        return image


def plot_tensor(image: Tensor, output: ndarray, plot_image=True, return_img=False):
    nimg = image[0].permute(1, 2, 0) * 255
    nimg = nimg.cpu().numpy().astype(np.uint8)
    nimg = cv2.cvtColor(nimg, cv2.COLOR_RGB2BGR)
    for idx in range(output.shape[0]):
        plot_skeleton_kpts(nimg, output[idx, 7:].T, 3)

    if plot_image:
        plt.figure(figsize=(8, 8))
        plt.axis("off")
        plt.imshow(nimg)
        plt.show()

    if return_img:
        return nimg



from aipose.frame.frame_manager_base import FrameManagerBase
from aipose.frame.frame_plot import FramePlot
from aipose.frame.frame_yolo_v7 import FrameYoloV7

__all__ = ["FrameManagerBase", "FramePlot", "FrameYoloV7"]



from typing import List

import torch
from numpy import ndarray

from aipose.frame import FrameManagerBase
from aipose.models.yolov7.domain import YoloV7Pose, YoloV7PoseKeypoints


class FrameYoloV7(FrameManagerBase):
    """
    A subclass of FrameManagerBase that uses YOLOv7 model to make predictions on each frame.

    Attributes:
        model: The YOLOv7 model used for predictions.

    Methods:
        __init__(): Initializes a new instance of the FrameYoloV7 class and sets the model to YOLOv7Pose.
        before_read_frame(): Overrides the before_read_frame method of FrameManagerBase, empties the GPU cache before reading the next frame.
        frame_received(frame: ndarray) -> ndarray: Overrides the frame_received method of FrameManagerBase, applies the model to the frame and processes the result.
        on_predict(frame: ndarray, prediction: List[YoloV7PoseKeypoints]) -> None | ndarray:
            This method is called after the prediction is made, can be overridden in the subclass for additional processing of the result.
    """

    def __init__(self):
        """
        Initializes a new instance of the FrameYoloV7 class and sets the model to YOLOv7Pose.
        """
        self.model = YoloV7Pose()

    def before_read_frame(self):
        """
        Overrides the before_read_frame method of FrameManagerBase, empties the GPU cache before reading the next frame.
        """
        torch.cuda.empty_cache()

    def frame_received(self, frame: ndarray) -> ndarray:
        """
        Overrides the frame_received method of FrameManagerBase, applies the model to the frame and processes the result.

        Args:
            frame (ndarray): The frame to process.

        Returns:
            ndarray: The processed frame.
        """
        prediction = self.model(frame)
        processed_frame = self.on_predict(frame, prediction)
        if processed_frame is None:
            return frame
        else:
            return processed_frame

    def on_predict(
        self,
        frame: ndarray,
        prediction: List[YoloV7PoseKeypoints],
    ) -> None | ndarray:
        """
        This method is called after the prediction is made, can be overridden in the subclass for additional processing of the result.

        Args:
            frame (ndarray): The input frame.
            prediction (List[YoloV7PoseKeypoints]): The predicted keypoints.

        Returns:
            None | ndarray: If None, the processed frame is not returned and the original frame is returned instead.
            If an ndarray is returned, it should be the processed frame.
        """
        raise NotImplementedError



from typing import List

import cv2
import numpy as np
from numpy import ndarray

from aipose.frame import FrameManagerBase
from aipose.models.yolov7.domain import YoloV7PoseKeypoints
from aipose.plot import plot


class FramePlot(FrameManagerBase):
    """
    A subclass of FrameManagerBase that plots the predictions of a given model on each frame.

    Attributes:
        model: The model used to generate the predictions.

    Methods:
        __init__(model): Initializes a new instance of the FramePlot class with the specified model.
        frame_received(frame: ndarray) -> ndarray: Overrides the frame_received method of FrameManagerBase, applies the model to the frame and plots the predictions.
        _plot(prediction: List[YoloV7PoseKeypoints], frame: ndarray) -> ndarray: Plots the predicted keypoints on the input image.
    """

    def __init__(self, model):
        """
        Initializes a new instance of the FramePlot class with the specified model.

        Args:
            model: The model used to generate the predictions.
        """
        self.model = model

    def frame_received(self, frame: ndarray) -> ndarray:
        """
        Overrides the frame_received method of FrameManagerBase, applies the model to the frame and plots the predictions.

        Args:
            frame (ndarray): The frame to process.

        Returns:
            ndarray: The modified frame.
        """
        yolo_v7_keypoints = self.model(frame)  # TODO: fix this
        frame = self._plot(yolo_v7_keypoints, frame)
        return frame

    def _plot(self, prediction: List[YoloV7PoseKeypoints], frame: ndarray) -> ndarray:
        """
        Plots the predicted keypoints on the input image.

        Args:
            prediction (List[YoloV7PoseKeypoints]): The predicted keypoints.
            frame (ndarray): The input image.

        Returns:
            ndarray: The image with the predicted keypoints plotted.
        """
        frame = plot(
            frame,
            np.array([value.raw_keypoint for value in prediction]),
            plot_image=False,
            return_img=True,
        )
        return frame



import hashlib
import os
from pathlib import Path

import requests
from tqdm import tqdm


class Downloader:
    """A class to download files from the internet."""

    def download(
        self,
        model_name: str,
        model_hash: str,
        url_download: str,
    ) -> str:
        """Download a file from the provided URL and store it locally.

        Args:
            model_name (str): Name of the model.
            model_hash (str): Hash of the expected model file.
            url_download (str): URL from where to download the model.

        Returns:
            str: Path to the downloaded model file.
        """
        aipose_path = Path.home() / ".aipose"
        os.makedirs(aipose_path, exist_ok=True)

        aipose_model_folder_path = aipose_path / model_name

        if not aipose_model_folder_path.is_file():
            self._download_file(url_download, aipose_model_folder_path)

        current_model_hash = self._calculate_hash(aipose_model_folder_path)

        if model_hash != current_model_hash:
            self._download_file(url_download, aipose_model_folder_path)

        return str(aipose_model_folder_path)

    def _calculate_hash(self, file_path: Path) -> str:
        """Calculate the MD5 hash of a file.

        Args:
            file_path (Path): Path to the file for which to calculate the hash.

        Returns:
            str: The MD5 hash of the file.
        """
        with file_path.open("rb") as f:
            return hashlib.md5(f.read()).hexdigest()

    def _download_file(self, url: str, file_path: Path) -> None:
        """Download a file from the provided URL and store it locally.

        Args:
            url (str): URL from where to download the file.
            file_path (Path): Path to save the downloaded file.

        Returns:
            None
        """
        with requests.get(url, stream=True) as r:
            r.raise_for_status()

            total_size = int(r.headers.get("Content-Length", 0))

            with file_path.open("wb") as f, tqdm(
                total=total_size, unit="B", unit_scale=True
            ) as pbar:
                for chunk in r.iter_content(chunk_size=8192):
                    f.write(chunk)
                    pbar.update(len(chunk))
